"""
A port of most of the globus-cli output formatter
"""

import json
from typing import Any, Callable, Iterable, Mapping, Optional, Tuple, Union, overload

import click

import globus_sdk
from globus_sdk.paging import Paginator

from gcs_cli.parsing.stateful_options import outformat_is_text

FORMAT_JSON = "json"
FORMAT_TEXT_TABLE = "text_table"
FORMAT_TEXT_RECORD = "text_record"
FORMAT_TEXT_RAW = "text_raw"

FieldSpec = Iterable[Tuple[str, Union[str, Callable]]]
ResponseDataType = Union[
    Paginator, Mapping, Iterable[Mapping], globus_sdk.GlobusHTTPResponse, str
]


def _key_to_keyfunc(k: Union[str, Callable]) -> Callable:
    """
    We allow for 'keys' which are functions that map columns onto value
    types -- they may do formatting or inspect multiple values on the
    object. In order to support this, wrap string keys in a simple function
    that does the natural lookup operation, but return any functions we
    receive as they are.
    """
    # if the key is a string, then the "keyfunc" is just a basic lookup
    # operation -- return that
    if isinstance(k, str):
        subkeys = k.split(".")

        def resolve_subkeys(x: Mapping) -> Mapping:
            container = x
            for sub in subkeys:
                container = container[sub]
            return container

        return resolve_subkeys
    # otherwise, the key must be a function which is executed on the item
    # to produce a value -- return it verbatim
    return k


def print_json_response(res: ResponseDataType) -> None:
    if isinstance(res, globus_sdk.GlobusHTTPResponse):
        res = res.data
    response = json.dumps(res, indent=2, separators=(",", ": "), sort_keys=True)
    click.echo(response)


def colon_formatted_print(data: Iterable, named_fields: FieldSpec) -> None:
    maxlen = max(len(n) for n, f in named_fields) + 1
    for name, field in named_fields:
        field_keyfunc = _key_to_keyfunc(field)
        click.echo("{} {}".format((name + ":").ljust(maxlen), field_keyfunc(data)))


def print_table(
    iterable: Iterable,
    headers_and_keys: FieldSpec,
    print_headers: bool = True,
) -> None:
    # the iterable may not be safe to walk multiple times, so we must walk it
    # only once -- however, to let us write things naturally, convert it to a
    # list and we can assume it is safe to walk repeatedly
    iterable = list(iterable)

    # extract headers and keys as separate lists
    headers = [h for (h, k) in headers_and_keys]
    keys = [k for (h, k) in headers_and_keys]

    # convert all keys to keyfuncs
    keyfuncs = [_key_to_keyfunc(key) for key in keys]

    # use the iterable to find the max width of an element for each column, in
    # the same order as the headers_and_keys array
    # use a special function to handle empty iterable
    def get_max_colwidth(kf: Callable) -> int:
        def _safelen(x: Any) -> int:
            try:
                return len(x)
            except TypeError:
                return len(str(x))

        lengths = [_safelen(kf(i)) for i in iterable]
        if not lengths:
            return 0
        else:
            return max(lengths)

    widths = [get_max_colwidth(kf) for kf in keyfuncs]
    # handle the case in which the column header is the widest thing
    widths = [max(w, len(h)) for w, h in zip(widths, headers)]

    # create a format string based on column widths
    format_str = " | ".join("{:" + str(w) + "}" for w in widths)

    @overload
    def none_or_bool_to_string(val: bool) -> str:
        ...

    @overload
    def none_or_bool_to_string(val: None) -> str:
        ...

    def none_or_bool_to_string(val: Any) -> Any:
        if val is None:
            return "NULL"
        elif isinstance(val, bool):
            return str(val)
        return val

    # print headers
    if print_headers:
        click.echo(format_str.format(*[h for h in headers]))
        click.echo(format_str.format(*["-" * w for w in widths]))
    # print the rows of data
    for i in iterable:
        click.echo(
            format_str.format(*[none_or_bool_to_string(kf(i)) for kf in keyfuncs])
        )


def formatted_print(
    response_data: ResponseDataType,
    simple_text: Optional[str] = None,
    text_preamble: Optional[str] = None,
    text_epilog: Optional[str] = None,
    text_format: str = FORMAT_TEXT_TABLE,
    json_converter: Callable[[ResponseDataType], Mapping] = None,
    fields: FieldSpec = None,
    page_callback: Callable[[globus_sdk.GlobusHTTPResponse], None] = None,
) -> None:
    """
    A generic output formatter. Consumes the following pieces of data:

    ``response_data`` is a dict or GlobusHTTPResponse object. It contains either an
    API response or synthesized data for printing.

    ``simple_text`` is a text override -- normal printing is skipped and this
    string is printed instead (text output only)
    ``text_preamble`` is text which prints before normal printing (text output
    only)
    ``text_epilog`` is text which prints after normal printing (text output
    only)
    ``text_format`` is one of the FORMAT_TEXT_* constants OR a callable which
    takes ``response_data`` and prints output. Note that when a callable is
    given, it does the actual printing

    ``json_converter`` is a callable that does preprocessing of JSON output. It
    must take ``response_data`` and produce another dict or dict-like object
    (json output only)

    ``fields`` is an iterable of (fieldname, keyfunc) tuples. When keyfunc is
    a string, it is implicitly converted to `lambda x: x[keyfunc]` (text output
    only)

    ``page_callback`` is a callback called on each page of results. On non-paginated
    data it is not called.
    """

    def page_iterator(data: Paginator) -> Iterable[globus_sdk.GlobusHTTPResponse]:
        for page in data.pages():
            if page_callback:
                page_callback(page)
            yield page

    def _assert_fields() -> None:
        if fields is None:
            raise ValueError(
                "Internal Error! Output format requires fields; none given. "
                "You can workaround this error by using `--format JSON`"
            )

    def _print_as_text() -> None:
        nonlocal response_data
        # unwrap paginated data into a generator which iterates over the contents of
        # each page for table text output
        # any non-table output on a paginated response is invalid!
        if isinstance(response_data, Paginator):
            if text_format != FORMAT_TEXT_TABLE:
                raise NotImplementedError(
                    "Internal misuse! paginated=True and format != table"
                )
            response_data = (
                x for page in page_iterator(response_data) for x in page.data["data"]
            )
        # if we're given simple text, print that and exit
        if simple_text is not None:
            click.echo(simple_text)
            return

        # if there's a preamble, print it beofre any other text
        if text_preamble is not None:
            click.echo(text_preamble)

        #  do the various kinds of printing
        if text_format == FORMAT_TEXT_TABLE:
            _assert_fields()
            if not isinstance(response_data, Iterable):
                raise ValueError(
                    "Internal Error! Output format requires iterable data. "
                    "You can workaround this error by using `--format JSON`"
                )
            assert isinstance(response_data, Iterable)
            assert fields is not None
            print_table(response_data, fields)
        elif text_format == FORMAT_TEXT_RAW:
            click.echo(response_data)
        elif text_format == FORMAT_TEXT_RECORD:
            _assert_fields()
            if isinstance(response_data, globus_sdk.GlobusHTTPResponse):
                response_data = response_data.data
            if not isinstance(response_data, Iterable):
                raise ValueError(
                    "Internal Error! Output format requires iterable data. "
                    "You can workaround this error by using `--format JSON`"
                )
            assert isinstance(response_data, Iterable)
            assert fields is not None
            colon_formatted_print(response_data, fields)

        # if there's an epilog, print it after any text
        if text_epilog is not None:
            click.echo(text_epilog)

    if outformat_is_text():
        _print_as_text()
    else:
        if json_converter:
            response_data = json_converter(response_data)
        # convert paginated data to an array
        if isinstance(response_data, Paginator):
            response_data = [x.data for x in page_iterator(response_data)]
        print_json_response(response_data)
