"""
This module contains interfaces and abstract base classes to help implement
code to check if a model is ready to migrate.
"""
from abc import abstractmethod, ABCMeta
from typing import (
    Callable,
    ClassVar,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)


T = TypeVar("T")


class Check(Generic[T], metaclass=ABCMeta):
    """
    This class is an abstract class if you need to implement additional checkers.
    """

    object_name: ClassVar[str]
    """
    The name of the object type being checked
    """

    @abstractmethod
    def is_ok(self, obj: Optional[T]) -> bool:
        """
        Method which is invoked by the CheckInvoker to cause the check to be
        run on a particular model. This may be invoked with None if the model
        is not defined, so that having required models can be enforced.

        The is_ok() method should be very tightly focused, so that we can
        provide clearer guidance should the check fail.

        This method returns True if the object passes the check, or False
        otherwise.
        """
        ...

    @property
    def name(self) -> str:
        """
        Name of the check. This should be of the form XXYY where XX is some
        indication of the object type (eg "GC" for guest collection) and YY
        is an integer for the error. We will be displaying these initially
        as part of the check diagnostics, but keeping these short, unique,
        and tied to specific models should help us document and hopefully
        automatically guide handling the check failure situations.
        """
        return self.__class__.__name__

    @property
    @abstractmethod
    def diagnostic(self) -> Optional[str]:
        """
        Diagnostic message that is printed when the check fails. For the
        initial built-in checks, this is called once per check type, and
        the Check-derived class keeps track of which models had the problem
        so that they can be included in the diagnostic message, so that
        the diagnostic output is less repetitive.
        """
        ...


class CheckInvoker(Generic[T], metaclass=ABCMeta):
    """
    This is a type-annotated version of a class which invokes all of the Check
    classes which are defined for the given type. Decorators below provide
    support for hooking checks into this object.
    """

    checkers: ClassVar[List]
    """
    List of Check classes associated with this type.
    """
    object_name: ClassVar[str]

    def __init__(self) -> None:
        self._checkers = [c() for c in self.__class__.checkers]

    def check(self, obj: Optional[T]) -> List[Check[T]]:
        """
        This method runs all of the checkers defined in the checkers member
        variable. Any checks which fail are added to the internal violations
        list, and the checks which failed for this object (if any) are returned.
        """
        violations: List[Check] = []

        for checker in self._checkers:
            try:
                checker_res = checker.is_ok(obj)
                if not checker_res:
                    violations.append(checker)
            except Exception:
                violations.append(checker)

        return violations

    @classmethod
    def checker(cls, target: Type[Check[T]]) -> Type[Check[T]]:
        class _Decorated(target):  # type: ignore
            object_name = cls.object_name

            @property
            def name(self) -> str:
                return target.__name__

        cls.checkers.append(_Decorated)

        return _Decorated

    @overload
    @classmethod
    def property_checker(cls, target: Type[Check[T]]) -> Type[Check[T]]:
        ...

    @overload  # noqa: F811
    @classmethod
    def property_checker(  # noqa: F811
        cls, *, property_name: str
    ) -> Callable[..., Type[Check[T]]]:
        ...

    @classmethod  # noqa: F811
    def property_checker(  # noqa: F811
        cls, target: Type[Check[T]] = None, property_name: str = None
    ) -> Union[Type[Check[T]], Callable[..., Type[Check[T]]]]:

        pn = property_name

        def decorator(target: Type[Check[T]]) -> Type[Check[T]]:
            class _Decorated(target):  # type: ignore
                object_name = cls.object_name
                property_name = pn

                @property
                def name(self) -> str:
                    return target.__name__

            cls.checkers.append(_Decorated)
            return _Decorated

        if target is None:
            return decorator
        else:
            return decorator(target)


class ObjectIsNotNoneCheck(Check[T]):
    """
    Concrete implementation of Check that validates that the object is not None
    """

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        self.violations = False

    def is_ok(self, obj: Optional[T]) -> bool:
        ok = True
        if not obj:
            ok = False
            self.violations = True
        return ok

    @property
    def diagnostic(self) -> Optional[str]:
        if self.violations:
            return "The object is null"
        return None


class PropertyCheck(Check[T]):
    """
    Abstract class for a Check object that checks a single property value.
    This invokes the abstract value_is_ok() method to check the attribute.
    This is passed the object being checked (not the property) in case the
    check implementation needs to check relationship between the checked
    property and some other state of the object.
    """

    property_name: ClassVar[str]
    """
    Each PropertyCheck implementation sets the property_naem to the name of the
    property being checked.
    """

    def __init__(self) -> None:
        self.violations: List[T] = []

    def is_ok(self, obj: Optional[T]) -> bool:
        """
        Concrete implementation of is_ok() which calls value_is_ok() if the
        object is not None.
        """
        ok = True
        if obj:
            if not self.value_is_ok(obj):
                if obj not in self.violations:
                    self.violations.append(obj)
                ok = False
        return ok

    @abstractmethod
    def value_is_ok(self, obj: T) -> bool:
        ...


class MissingPropertyCheck(PropertyCheck[T]):
    """
    Concrete implementation of PropertyCheck that catches objects that
    have the checked property set to None.
    """

    def value_is_ok(self, value: T) -> bool:
        ok = True
        if getattr(value, self.property_name) is None:
            ok = False
        return ok

    @property
    def diagnostic(self) -> Optional[str]:
        if self.violations:
            violation_list = [f"{v}" for v in self.violations]
            if len(violation_list) > 0:
                return (
                    f"The {self.property_name} property is not set on the {self.object_name}:\n    "
                    + "\n    ".join(violation_list)
                )
        return None


class PropertyConstraintCheck(PropertyCheck[T]):
    """
    Abstract class for a PropertyCheck that calls value_is_ok when the
    attribute is defined on the object and has a value other than None.
    """

    constraint_name: ClassVar[str]
    """
    Name of the constraint that is being checked. This is included in the
    diagnostic message
    """

    def is_ok(self, obj: Optional[T]) -> bool:
        ok = True
        if obj:
            if (
                hasattr(obj, self.property_name)
                and getattr(obj, self.property_name) is not None
            ):
                if not self.value_is_ok(obj):
                    if obj not in self.violations:
                        self.violations.append(obj)
                    ok = False
        return ok

    @property
    def diagnostic(self) -> Optional[str]:
        if self.violations:
            violation_list = [f"{v}" for v in self.violations]
            if len(violation_list):
                return (
                    f"The {self.property_name} property value is not valid "
                    f"({self.constraint_name}) for the {self.object_name}:\n    "
                    + "\n    ".join(violation_list)
                )
        return None


U = TypeVar("U")


class RelationshipChecker(Generic[T, U], Check[U]):
    """
    Implementation of Check that checks relationships between multiple
    objects, so that either could be fixed to eliminate the violation
    """

    relationship_object_names: ClassVar[List[str]]
    constraint_name: ClassVar[str]

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        self.violations: List[List[T]] = []

    def is_ok(self, obj: Optional[U]) -> bool:
        ok = True
        err_objs: List[List[T]] = []
        if obj is not None:
            err_objs = self.relationship_is_ok(obj)
        if err_objs:
            for err_obj in err_objs:
                if err_obj not in self.violations:
                    self.violations.append(err_obj)
            ok = False
        return ok

    @abstractmethod
    def relationship_is_ok(self, obj: U) -> List[List[T]]:
        ...

    @property
    def relationship_object_name_list(self) -> str:
        if len(self.relationship_object_names) < 2:
            return str(self.relationship_object_names)
        elif len(self.relationship_object_names) == 2:
            return " and ".join(self.relationship_object_names)
        else:
            return (
                ", ".join(self.relationship_object_names[:-1])
                + ", and "
                + self.relationship_object_names[-1]
            )

    @property
    def diagnostic(self) -> Optional[str]:
        if self.violations:
            violation_strings: List[str] = []
            for v in self.violations:
                violation_str = ", ".join([str(w) for w in v])
                violation_strings.append(f"({violation_str})")
            return (
                f'The relationship "{self.constraint_name}" for '
                f"{self.relationship_object_name_list} is invalid for "
                f"{violation_strings}"
            )
        return None
