import datetime
import json
import os
import pathlib
import stat
from tempfile import TemporaryDirectory
from typing import Any, Dict, Optional
from zipfile import ZipFile, ZIP_DEFLATED, ZIP_STORED

import click

from ..version import __version__
from gcs_migration.models import MigrationPlan
from .decorators import requires_migration_plan


@click.command("create")
@click.version_option(version=__version__)
@click.option(
    "--description",
    "-d",
    metavar="DESCRIPTION",
    help="Text description of the snapshot",
)
@click.option("-o", "--omit-secrets", is_flag=True, help="Omit secrets from snapshot")
@requires_migration_plan
def snapshot_create(
    description: Optional[str], migration_plan: MigrationPlan, omit_secrets: bool
) -> None:
    os.umask(stat.S_IRWXG | stat.S_IRWXO)
    snapshot_dir = pathlib.Path("./migration_snapshots")
    if not snapshot_dir.exists():
        os.makedirs(snapshot_dir, mode=0o700)

    snapshot_number = 1

    for fn in os.listdir(snapshot_dir):
        try:
            idx = int(fn)
            if idx >= snapshot_number:
                snapshot_number = idx + 1
        except ValueError:
            pass

    dt = datetime.datetime.now()
    dt = dt - datetime.timedelta(microseconds=dt.microsecond)

    metadata: Dict[str, Any] = {"timestamp": str(dt)}
    if description:
        metadata["description"] = description
    metadata["omit_secrets"] = omit_secrets

    with TemporaryDirectory() as tmpdir:
        tmpdir_path = pathlib.Path(tmpdir)
        migration_plan.root = tmpdir_path
        migration_plan.dump(omit_secrets=omit_secrets)

        with ZipFile(
            snapshot_dir / f"{snapshot_number:04}", mode="w", compression=ZIP_DEFLATED
        ) as z:
            z.writestr("metadata.json", json.dumps(metadata), ZIP_STORED)
            for root, _dirs, files in os.walk(tmpdir_path / "migration_plan"):
                root_path = pathlib.Path(root)
                for fn in files:
                    z.write(
                        root_path / fn,
                        arcname=(root_path / fn).relative_to(tmpdir_path),
                    )
    click.echo(f"Created snapshot {snapshot_number}")
