from typing import Any, cast, Dict, List, Optional
from typing_extensions import TypedDict

from ..parsers import InCommonMetadata, CalloutDict
from ..util import yaml, yaml_object


class CILogonConfigurationDict(TypedDict):
    organization: Optional[str]
    domains: Optional[List[str]]
    dn: Optional[str]
    callout: Optional[CalloutDict]


@yaml_object(yaml)
class CILogonConfiguration:
    def __init__(self, callout: Optional[CalloutDict] = None) -> None:
        self.domains: Optional[List[str]] = None
        self.organization: Optional[str] = None
        self.dn: Optional[str] = None
        self.callout = callout

    @property
    def is_myproxy(self) -> bool:
        return False

    @property
    def is_cilogon(self) -> bool:
        return True

    @property
    def is_gridmap(self) -> bool:
        return False

    @property
    def callout(self) -> Optional[CalloutDict]:
        return self._callout

    @callout.setter
    def callout(self, value: Optional[CalloutDict]) -> None:
        self._callout = value
        if value and not self.dn:
            myproxy_dn = value["environment"].get("GLOBUS_MYPROXY_AUTHORIZED_DN")
            if myproxy_dn:
                assert isinstance(myproxy_dn, str)
                self.dn = myproxy_dn

    @property
    def dn(self) -> Optional[str]:
        return self._dn

    @dn.setter
    def dn(self, value: Optional[str]) -> None:
        self._dn = value
        if value and not self.organization:
            self.organization = value.rsplit("/O=", 1)[1]

    @property
    def organization(self) -> Optional[str]:
        return self._organization

    @organization.setter
    def organization(self, value: Optional[str]) -> None:
        self._organization = value
        if value and not self.domains:
            metadata = InCommonMetadata(organization=value)
            if metadata:
                self.domains = metadata.domains

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CILogonConfiguration):
            return NotImplemented
        if self is other:
            return True

        self_state = cast(Dict[str, Any], self.__getstate__())
        self_state.pop("domains", None)

        other_state = cast(Dict[str, Any], other.__getstate__())
        other_state.pop("domains", None)

        return self_state == other_state

    def __getstate__(self) -> CILogonConfigurationDict:
        return {
            "organization": self.organization,
            "domains": self.domains,
            "dn": self.dn,
            "callout": self.callout,
        }

    def __setstate__(self, state: CILogonConfigurationDict) -> None:
        self.domains = state["domains"]
        self.organization = state["organization"]
        self.dn = state["dn"]
        self.callout = state["callout"]
