import os
import pathlib
from typing import Dict, Optional, List, Union
from typing_extensions import TypedDict

from .cilogon_configuration import CILogonConfiguration
from .gridmap_configuration import GridmapConfiguration
from .myproxy_configuration import MyProxyConfiguration
from .unsupported_callout_configuration import UnsupportedCalloutConfiguration
from ..parsers import CalloutDict, GsiAuthzParser, GridFTPOptionParser
from ..util import yaml_object, yaml


CalloutConfiguration = Union[
    CILogonConfiguration,
    GridmapConfiguration,
    MyProxyConfiguration,
    UnsupportedCalloutConfiguration,
]


class GsiAuthzConfigurationDict(TypedDict):
    callouts: Optional[List[CalloutConfiguration]]
    gsi_authz_conf: Optional[str]
    gridmap: Optional[str]


@yaml_object(yaml)
class GsiAuthzConfiguration:
    def __init__(
        self,
        gridmap: "Optional[Union[str, os.PathLike[str]]]" = None,
        gsi_authz_conf: "Optional[Union[str, os.PathLike[str]]]" = None,
        gridftp_config: Optional[GridFTPOptionParser] = None,
    ) -> None:
        self.callout_dict_list: Optional[List[Optional[CalloutDict]]] = None
        self.callouts: Optional[List[CalloutConfiguration]] = None
        self.gsi_authz_conf: "Optional[Union[str, os.PathLike[str]]]" = gsi_authz_conf
        self.gridmap: "Optional[Union[str, os.PathLike[str]]]" = gridmap
        self.environment: "Dict[str, str]" = {}

        parser = GsiAuthzParser()
        self.callout_dict_list = parser.parse(gsi_authz_conf, gridmap, gridftp_config)
        if gridftp_config:
            self.environment = gridftp_config.get_environment()

    @property
    def callout_dict_list(self) -> Optional[List[Union[CalloutDict, None]]]:
        return self._callout_dict_list

    @callout_dict_list.setter
    def callout_dict_list(
        self, value: Optional[List[Union[CalloutDict, None]]]
    ) -> None:
        self._callout_dict_list = value

        if value:
            callouts: List[CalloutConfiguration] = []
            for c in value:
                if value and self.__is_cilogon(c):
                    callouts.append(CILogonConfiguration(callout=c))
                elif value and self.__is_myproxy(c):
                    callouts.append(MyProxyConfiguration(callout=c))
                elif value and self.__is_unsupported(c):
                    callouts.append(UnsupportedCalloutConfiguration(callout=c))
                else:
                    callouts.append(GridmapConfiguration(gridmap=self.gridmap))
            self.callouts = callouts

    @staticmethod
    def __is_myproxy(callout: Optional[CalloutDict]) -> bool:
        if callout is None:
            return False
        else:
            library = callout.get("library")
            symbol = callout.get("symbol")

        return (
            library == "libglobus_gridmap_verify_myproxy_callout"
            and symbol == "globus_gridmap_verify_myproxy_callout"
        )

    @staticmethod
    def __is_cilogon(callout: Optional[CalloutDict]) -> bool:
        if callout is None:
            return False
        else:
            library = callout.get("library")
            symbol = callout.get("symbol")

        return (
            library == "libglobus_gridmap_eppn_callout"
            and symbol == "globus_gridmap_eppn_callout"
        )

    @staticmethod
    def __is_unsupported(callout: Optional[CalloutDict]) -> bool:
        if callout is None:
            return False
        else:
            library = callout.get("library")

        return bool(
            library
            and library
            not in [
                "libglobus_gridmap_eppn_callout",
                "libglobus_gridmap_verify_myproxy_callout",
            ]
        )

    @property
    def cilogon_callouts(self) -> List[CILogonConfiguration]:
        if self.callouts:
            return [c for c in self.callouts if isinstance(c, CILogonConfiguration)]
        else:
            return []

    @property
    def myproxy_callouts(self) -> List[MyProxyConfiguration]:
        if self.callouts:
            return [c for c in self.callouts if isinstance(c, MyProxyConfiguration)]
        else:
            return []

    @property
    def gridmap_configuration(self) -> List[GridmapConfiguration]:
        if self.callouts:
            return [c for c in self.callouts if isinstance(c, GridmapConfiguration)]
        else:
            return []

    def dump(self, root: "Union[str, os.PathLike[str]]") -> None:
        filename = pathlib.Path(root) / "gsi_authz_configuration.yaml"
        with open(filename, "w") as f:
            yaml.dump(self, f)

    @staticmethod
    def load(
        root: "Union[str, os.PathLike[str]]", load_version: str
    ) -> Optional["GsiAuthzConfiguration"]:
        filename = pathlib.Path(root) / "gsi_authz_configuration.yaml"
        if os.path.exists(filename):
            with open(filename, "r") as f:
                obj = yaml.load(f)
                assert isinstance(obj, GsiAuthzConfiguration)

                return obj
        return None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GsiAuthzConfiguration):
            return NotImplemented
        return self is other or self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> GsiAuthzConfigurationDict:
        if isinstance(self.gsi_authz_conf, os.PathLike):
            gsi_authz_conf: Optional[str] = self.gsi_authz_conf.__fspath__()
        else:
            gsi_authz_conf = self.gsi_authz_conf
        if isinstance(self.gridmap, os.PathLike):
            gridmap: Optional[str] = self.gridmap.__fspath__()
        else:
            gridmap = self.gridmap

        return {
            "callouts": self.callouts,
            "gsi_authz_conf": gsi_authz_conf,
            "gridmap": gridmap,
        }

    def __setstate__(self, state: GsiAuthzConfigurationDict) -> None:
        self.gsi_authz_conf = state["gsi_authz_conf"]
        self.gridmap = state["gridmap"]
        self.callouts = state["callouts"]
        self._callout_dict_list = None  # This is not serialized
