import json
import os
from typing import Any, cast, Dict

from ..typing.gcs_documents import (
    ExpressionIdentityMappingDict,
    ExpressionIdentityMappingEntryDict,
    ExternalIdentityMappingDict,
    IdentityMappingDict,
)


def parse_identity_mapping(idm: str) -> IdentityMappingDict:
    if os.path.exists(idm):
        with open(idm, "r") as f:
            mapping = json.load(f)
    else:
        mapping = json.loads(idm)

    if not isinstance(mapping, dict):
        raise KeyError("Error parsing identity mapping")

    if "DATA_TYPE" not in mapping:
        raise KeyError("Missing required key DATA_TYPE in identity mapping")

    if mapping["DATA_TYPE"] == "expression_identity_mapping#1.0.0":
        _check_keys(
            mapping, ExpressionIdentityMappingDict, "expression_identity_mapping"
        )
        for entry in mapping["mappings"]:
            _check_keys(
                entry,
                ExpressionIdentityMappingEntryDict,
                "expression_identity_mapping entry",
            )
        return cast(ExpressionIdentityMappingDict, mapping)
    elif mapping["DATA_TYPE"] == "external_identity_mapping#1.0.0":
        _check_keys(mapping, ExternalIdentityMappingDict, "external_identity_mapping")
        return cast(ExternalIdentityMappingDict, mapping)
    else:
        raise ValueError("Invalid DATA_TYPE in identity mapping")


def _check_keys(mapping: Dict[str, Any], schema: Any, schema_name: str) -> None:
    # disable type checking on __required_keys__ and __optional_keys__ for
    # https://github.com/python/mypy/issues/13358
    for required in schema.__required_keys__:  # type: ignore
        if required not in mapping:
            raise KeyError(f"Missing required key {required} in {schema_name}")
    for key in mapping.keys():
        if key not in schema.__required_keys__ and key not in schema.__optional_keys__:  # type: ignore
            raise KeyError(f"Unexpected key {key} in {schema_name}")
