import collections
import datetime
import logging
import logging.handlers
import os
import pkg_resources
import random
import string
from typing import Optional, Union
import uuid

from flask import (
    Flask,
    g,
    jsonify,
    make_response,
    render_template,
    request,
)

from flask_cors import CORS
from werkzeug.exceptions import NotFound

from . import blueprints as blueprints
from globus.manager.constants import NodeStatus
from globus.manager.utils.keychain import KeyChain
from globus.manager.api.errors import (
    APIBaseError,
    RouteNotFoundError,
    ServiceUnavailableError,
)
import globus.manager.api.utils.services as services
from globus.manager.api.schemas.api import result_schema
from globus.manager.api.models import GlobusDnsDomain, Endpoint, Node


class GCSManagerApp(Flask):
    def __init__(self, name):
        self._appsync_client = None
        self._auth_client = None
        self._dns_client = None
        self._transfer_client = None
        self._search_client = None
        self.node_status = NodeStatus.active
        self.node_status_last_check = datetime.datetime.fromtimestamp(
            0, datetime.timezone.utc
        )
        self.subscription_last_check = datetime.datetime.fromtimestamp(
            0, datetime.timezone.utc
        )
        self._keychain = KeyChain()
        self._gridftp_key_path = os.path.expanduser("~gcsweb/gridftp-key")
        self._gridftp_key_path_old = self._gridftp_key_path + ".old"
        self._gridftp_key = None
        self._gridftp_key_old = None
        self._gridftp_key_timestamp = None
        self._gridftp_key_old_timestamp = None

        super().__init__(name)

        self.register_error_handler(
            APIBaseError, self.handle_validation_error_as_result
        )

        self.register_error_handler(NotFound, self.handle_not_found_as_result)

        self.before_request(self.pre_request_g_setup)
        self.before_request(self.check_node_is_up)

        log_already_registered = False
        add_request_id_header_already_registered = False
        for _k, v in self.after_request_funcs:
            if self.log_response in v:
                log_already_registered = True
                continue
            if self.add_request_id_header in v:
                add_request_id_header_already_registered = True
                continue

        if not log_already_registered:
            self.after_request(self.log_response)
        if not add_request_id_header_already_registered:
            self.after_request(self.add_request_id_header)

    @property
    def node(self) -> Optional[Node]:
        domain = GlobusDnsDomain.query.one_or_none()
        my_node: Optional[Node] = None
        if domain:
            my_node = domain.get_node(self.config["NODE_ID"])
        return my_node

    def check_node_is_up(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        self.logger.debug("Cached node status is %s", str(self.node_status))
        # In some cases tests don't have all the expected data in the db
        if (now - self.node_status_last_check) > datetime.timedelta(minutes=1):
            self.logger.debug("Rechecking if our node is up")
            my_node = self.node
            if my_node:
                self.logger.debug("DB node status is %s", str(my_node.status))
                self.node_status = my_node.status
                self.node_status_last_check = now

        if self.node_status == NodeStatus.inactive:
            raise ServiceUnavailableError()

    def pre_request_g_setup(self):
        g.request_id = "".join(
            random.choice(string.ascii_letters + string.digits) for x in range(9)
        )
        g.created_objects = []
        g.updated_objects = []
        g.deleted_objects = []

    def add_request_id_header(self, response):
        response.headers["x-gcs-request-id"] = g.request_id
        response.headers["x-gcs-node-id"] = self.config["NODE_ID"]

        return response

    def log_response(self, response):
        log_items = collections.OrderedDict()

        request_id = getattr(g, "request_id", None)

        if request_id:
            log_items["request_id"] = request_id

        log_items["method"] = request.method
        log_items["path"] = request.path
        if response:
            log_items["status_code"] = response.status_code

        remote_addr = None

        remote_addr = request.environ.get("HTTP_X_FORWARDED_FOR")
        if remote_addr:
            remote_addr = remote_addr.split(",")[-1].strip()
        else:
            remote_addr = request.environ.get("REMOTE_ADDR")
        if remote_addr:
            log_items["client_address"] = remote_addr

        if hasattr(g, "authorized_identities"):
            authorized_identity_info = []
            for i in g.authorized_identities:
                this_info = collections.OrderedDict()
                this_info["id"] = i["id"]
                this_info["username"] = i["username"]
                if "_role_name" in i:
                    this_info["role"] = i["_role_name"]
                if "_role_id" in i:
                    this_info["role_id"] = i["_role_id"]
                authorized_identity_info.append(
                    ",".join(["{}:{}".format(k, v) for k, v in this_info.items()])
                )
            log_items["authorization"] = ",".join(authorized_identity_info)
        elif hasattr(g, "auth_state"):
            authenticated_identity_info = []
            for i in g.auth_state.identity_set_data:
                this_info = collections.OrderedDict()
                this_info["id"] = i["id"]
                this_info["username"] = i["username"]
                authenticated_identity_info.append(
                    ",".join(["{}:{}".format(k, v) for k, v in this_info.items()])
                )
            log_items["authorization"] = ",".join(authenticated_identity_info)
        else:
            log_items["authorization"] = "unauthenticated"

        if g.created_objects:
            log_items["created"] = ",".join(g.created_objects)

        if g.updated_objects:
            log_items["updated"] = ",".join(g.updated_objects)

        if g.deleted_objects:
            log_items["deleted"] = ",".join(g.deleted_objects)

        if response:
            response_json = response.get_json()
            if response_json:
                if "code" in response_json:
                    log_items["code"] = response_json["code"]
                elif "error" in response_json:
                    log_items["error"] = response_json["error"]

            self.logger.info(
                " ".join("{}={}".format(k, v) for k, v in log_items.items())
            )

        return response

    @staticmethod
    def handle_not_found_as_result(error):
        return GCSManagerApp.negotiate_error_result_response(
            request, RouteNotFoundError()
        )

    @staticmethod
    def handle_validation_error_as_result(error):
        return GCSManagerApp.negotiate_error_result_response(request, error)

    @staticmethod
    def negotiate_error_result_response(request, error):
        detail = error.detail
        if not detail:
            detail = str(error.errors)

        if "text/html" == request.accept_mimetypes.best:
            response = make_response(
                render_template(
                    "error.jinja2", status_code=error.status_code, **error.to_dict()
                )
            )
        else:
            res = result_schema.dump(
                {
                    "http_response_code": error.status_code,
                    "code": error.error,
                    "detail": detail,
                    "message": error.error_description,
                }
            )

            response = jsonify(res)
        response.status_code = error.status_code
        return response

    @property
    def client_id(self):
        return self.config["CLIENT_ID"]

    @property
    def client_secret(self):
        return self.config["CLIENT_SECRET"]

    @property
    def environment(self):
        environment = self.config["ENVIRONMENT"]
        if environment == "default":
            environment = "production"
        return environment

    @property
    def home(self) -> str:
        return self.config["HOME"]

    @property
    def config_key_file(self) -> str:
        return os.path.join(self.home, "gridftp-key")

    @property
    def vhost_cred_dir(self) -> str:
        return os.path.join(self.home, "etc/domains")

    @property
    def subscription(self) -> Optional[str]:
        now = datetime.datetime.now(datetime.timezone.utc)
        elapsed = now - self.subscription_last_check
        # In some cases tests don't have all the expected data in the db
        if (elapsed > datetime.timedelta(seconds=30)) or not self.config.get(
            "SUBSCRIPTION_ID"
        ):
            endpoint = Endpoint.query.one_or_none()
            if endpoint:
                new_subscription_id = getattr(endpoint, "subscription_id", None)
                if new_subscription_id != self.config.get("SUBSCRIPTION_ID"):
                    self.logger.debug(
                        "subscription_change_detected: subscription_id=%s",
                        new_subscription_id,
                    )
                    self.config["SUBSCRIPTION_ID"] = new_subscription_id
            self.subscription_last_check = now
        return self.config.get("SUBSCRIPTION_ID", None)

    @subscription.setter
    def subscription(self, value: Optional[Union[str, uuid.UUID]]) -> None:
        self.logger.debug("Setting subscription to %s", value)
        if isinstance(value, uuid.UUID):
            self.config["SUBSCRIPTION_ID"] = str(value)
        else:
            self.config["SUBSCRIPTION_ID"] = value

    @property
    def appsync_client(self):
        if not self._appsync_client:
            self._appsync_client = services.get_appsync_client(
                client_id=self.client_id,
                secret=self.client_secret,
                environment=self.environment,
            )
        return self._appsync_client

    @property
    def auth_client(self):
        if not self._auth_client:
            self._auth_client = services.get_auth_client(
                client_id=self.client_id,
                secret=self.client_secret,
                environment=self.environment,
            )
        return self._auth_client

    @property
    def dns_client(self):
        if not self._dns_client:
            self._dns_client = services.get_dns_client(
                client_id=self.client_id,
                secret=self.client_secret,
                environment=self.environment,
            )
        return self._dns_client

    @property
    def transfer_client(self):
        if not self._transfer_client:
            self._transfer_client = services.get_transfer_client(
                client_id=self.client_id,
                secret=self.client_secret,
                environment=self.environment,
            )
        return self._transfer_client

    @property
    def search_client(self):
        if not self._search_client:
            self._search_client = services.get_search_client(
                client_id=self.client_id,
                secret=self.client_secret,
                environment=self.environment,
            )
        return self._search_client

    @property
    def keychain(self):
        return self._keychain

    def _file_changed(self, path, timestamp):
        try:
            stat_res = os.stat(path)
        except (FileNotFoundError, IOError):
            return False
        return timestamp != stat_res.st_mtime

    @property
    def gridftp_keys(self):
        gridftp_keys = []

        if not self._gridftp_key_old or self._file_changed(
            self._gridftp_key_path_old, self._gridftp_key_old_timestamp
        ):
            try:
                with open(self._gridftp_key_path_old, "r") as f:
                    self._gridftp_key_old = f.read()
                    self._gridftp_key_old_timestamp = os.fstat(f.fileno()).st_mtime
            except (FileNotFoundError, IOError):
                pass

        if not self._gridftp_key or self._file_changed(
            self._gridftp_key_path, self._gridftp_key_timestamp
        ):
            try:
                with open(self._gridftp_key_path, "r") as f:
                    self._gridftp_key = f.read()
                    self._gridftp_key_timestamp = os.fstat(f.fileno()).st_mtime
            except (FileNotFoundError, IOError):
                pass

        if self._gridftp_key_old:
            gridftp_keys.append(self._gridftp_key_old)
        if self._gridftp_key:
            gridftp_keys.append(self._gridftp_key)

        return gridftp_keys


def make_app(**kwargs):
    logger = logging.getLogger()

    if logger.hasHandlers():
        default_handler = logger.handlers[0]
    else:
        default_handler = logging.StreamHandler()
        logger.addHandler(default_handler)

    app = GCSManagerApp(__name__)

    app.config.from_object("globus.manager.api.config.default")
    app.config.from_envvar("MANAGER_CONFIG", silent=True)

    if kwargs.get("app_config"):
        app.config.from_object(kwargs["app_config"])

    if not app.config.get("CLI"):
        # Turn on cross-origin support for this app
        CORS(app, max_age=600)

        # Turn of strict trailing slash enforcement in the underlying
        # Werkzeug routing engine.
        # With strict_slashes = False, http requests with trailing
        # slashes will silently match routes defined without the slash
        # e.g., GET /test/ will match @route('/test').
        app.url_map.strict_slashes = False

        for bp in blueprints.blueprints:
            app.register_blueprint(bp)

        # Load user mapping modules
        app.config["mapping"] = {}
        for entry_point in pkg_resources.iter_entry_points("globus.manager.mapping"):
            user_mapping = entry_point.load()
            app.config["mapping"][entry_point.name] = user_mapping

    if "LOG_HANDLER" in app.config:
        log_handler = app.config["LOG_HANDLER"]
        log_handler.setFormatter(
            logging.Formatter("%(asctime)s %(message)s", datefmt="%Y-%m-%dT%H:%M:%S%z")
        )
        app.logger.addHandler(log_handler)
        logger.removeHandler(default_handler)

    log_level = os.environ.get("GCS_MANAGER_LOG_LEVEL")
    if not log_level:
        log_level = app.config.get("GCS_MANAGER_LOG_LEVEL", "WARN")
    app.logger.setLevel(log_level)

    # Load connector modules, passing in the app.
    # They may register additional blueprints with the app.
    app.config["connectors"] = {}
    with app.app_context():
        for entry_point in pkg_resources.iter_entry_points("globus.manager.connectors"):
            connector = entry_point.load()(app)
            app.config["connectors"][connector.CONNECTOR_ID] = connector

    return app
