import functools
import re
from typing import Callable, Optional, overload, Type, TypeVar
from typing_extensions import ParamSpec, Protocol

import base62
from flask import current_app, g, request
from werkzeug.exceptions import HTTPException
from werkzeug.routing import BaseConverter

import marshmallow as ma
from globus.manager.api.errors import (
    BadRequestError,
    ValidationError,
    UnsupportedMediaError,
)
from globus.manager.api.utils.server_timing import ServerTiming


T = TypeVar("T")
P = ParamSpec("P")


@overload
def expects_json(fn: Callable[P, T]) -> Callable[P, T]:
    ...


@overload
def expects_json(
    *,
    schema: Optional[ma.Schema] = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


def expects_json(
    fn: Callable[P, T] = None,
    *,
    schema: Optional[ma.Schema] = None,
):
    """
    Decorate a route to require a JSON body and optionally validate against a
    marshmallow schema.

    :param fn:
        View method to execute if the expectation is met. When called, this
        decorator will populate `g.validated_data`. If there is no schema
        associated with this decorator, the value of `g.validated_data` is
        set to `None`, otherwise it is the output of deserializing the input
        object with the marshmallow schema.
    :param schema:
        A schema instance to use to validate the input.

    :raises UnsupportedMediaError:
        The input content-type is not application/json
    :raises BadRequestError:
        The input could not be loaded from the HTTP request.
    :raises ValidationError:
        The input does not validate against the schema.
    """
    if fn is None:
        return functools.partial(expects_json, schema=schema)

    @functools.wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        if not request.is_json:
            raise UnsupportedMediaError()

        try:
            json = request.get_json()
        except HTTPException as ex:
            raise BadRequestError(ex.description)

        g.validated_data = None

        if schema:
            try:
                g.validated_data = schema.load(json)
                current_app.logger.debug(
                    "Validated Data ({}): {}".format(
                        schema.__class__.__name__ if schema else "any",
                        g.validated_data,
                    )
                )
            except ma.ValidationError as e:
                current_app.logger.error(
                    "deserialization error: {} {} using schema {}".format(
                        e, json, schema
                    )
                )
                raise ValidationError(exc=e)

        assert fn is not None
        return fn(*args, **kwargs)

    return decorated_function


# Server metrics we want to provide to API consumers
db_read = ("db-read", "Database Reads")
db_write = ("db-write", "Database Writes")
transfer_api = ("t-api", "Transfer API")
auth_api = ("auth-api", "Auth API")
letsencrypt = ("letsencrypt", "LetsEncrypt")


def server_timing(fn: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        g.server_timing = ServerTiming()

        assert fn is not None
        return fn(*args, **kwargs)

    return decorated_function


class Base62Marker:
    """
    Generic safe string marker implementation. This is used as the
    default marker class for the paginated decorator.
    """

    def __init__(self, marker: str = None, marker_data: str = None) -> None:
        """
        Initializer for the Base62Marker class. One of the two named parameters
        can be used to initialize the marker.

        :param marker:
            The base62 encoded representation of this marker. This is typically
            used when parsing the marker query parameter to a route.
        :param marker_data:
            A string version of the data which will be stored in the marker.
        """
        if marker:
            self.decode(marker)
            if marker_data:
                raise ValueError("Invalid marker")
        elif marker_data:
            self.marker = marker_data
        else:
            raise ValueError("Missing required marker or marker_data")

    @property
    def marker(self) -> str:
        """
        The string version of the data stored in the marker.
        """
        return self._marker

    @marker.setter
    def marker(self, value: str) -> None:
        self._marker = value

    def decode(self, marker: str) -> None:
        """
        Decode the base62 marker string and store its value in the
        `marker` property
        """
        self.marker = base62.decodebytes(marker).decode("utf-8")

    def encode(self) -> str:
        """
        Encode and return the `marker` property as a base62 string.
        """
        return base62.encodebytes(self.marker.encode("utf-8"))

    def __str__(self) -> str:
        return self.marker


class Paginator(Protocol):
    def __init__(self, marker: str) -> None:
        ...

    def decode(self, marker: str) -> None:
        ...

    def encode(self) -> str:
        ...


@overload
def paginated(fn: Callable[P, T]) -> Callable[P, T]:
    ...


@overload
def paginated(
    *, marker_class: Type[Paginator] = None
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


def paginated(fn: Callable[P, T] = None, *, marker_class: Type[Paginator] = None):
    """
    Decorate a route to parse and validate pagination parameters `marker` and
    `page_size`. The results of parsing are stored in `g.marker` and
    `g.page_size`.

    :param fn:
        The route handler to call after parsing the optional pagination
        parameters.
    :param marker_class:
        An optional class that implements a constructor that can be passed a
        `marker` named parameter with encode() and decode() methods to parse
        and serialize the marker. If not present or None, the
        :py:class:`globus.manager.api.blueprints.decorators.Base62Marker` class
        is used.
    """
    if fn is None:
        return functools.partial(paginated, marker_class=marker_class)

    if marker_class is None:
        marker_class = Base62Marker

    @functools.wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        page_size_string = request.args.get("page_size", "16")
        marker_string = request.args.get("marker")
        page_size = None
        marker = None

        try:
            page_size = int(page_size_string)
            if page_size < 1 or page_size > 100:
                raise ValueError(
                    "The page_size parameter must be a positive value between "
                    "1 and 100"
                )
        except ValueError:
            raise BadRequestError("Invalid page_size parameter")

        if marker_string:
            try:
                assert marker_class is not None
                marker = marker_class(marker=marker_string)
            except (ValueError, AttributeError, UnicodeDecodeError):
                raise BadRequestError("Invalid marker")

        g.page_size = page_size
        g.marker = marker
        assert fn is not None
        return fn(*args, **kwargs)

    return decorated_function


@overload
def query_parameter(fn: Callable[P, T]) -> Callable[P, T]:
    ...


@overload
def query_parameter(
    *, name_spec: str = None, required: bool = False
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


def query_parameter(
    fn: Callable[P, T] = None, *, name_spec: str = None, required: bool = False
):
    """
    Decorate a route to parse and validate a query parameter using the
    same converters as flask uses for query parameters. The converted value
    is passed to the route using the name of parameter.

    :param fn:
        The route handler to call after parsing the optional pagination
        parameters.
    :param name_spec:
        A string containing the name of the query parameter or a string
        of the form <converter:name>, in which case the flask URL
        map converter will be invoked.
    :param required:
        Boolean flag indicating that this parameter is required. If it is omitted,
        a ValueError() exception is raised.
    """
    if fn is None:
        return functools.partial(
            query_parameter, name_spec=name_spec, required=required
        )

    @functools.wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        assert name_spec is not None
        match = re.match("<(.*):(.*)>", name_spec)
        converter: Optional[BaseConverter] = None
        if match:
            converter_name = match.group(1)
            name = match.group(2)
            converter = current_app.url_map.converters.get(converter_name)(
                current_app.url_map
            )
        else:
            name = name_spec

        raw_value = request.args.get(name)
        if raw_value is None and required:
            raise ValueError(
                "Missing required value for query parameter {}".format(name)
            )
        if converter:
            kwargs[name] = converter.to_python(raw_value)
        else:
            kwargs[name] = raw_value

        assert fn is not None
        return fn(*args, **kwargs)

    return decorated_function
