import uuid

from flask import current_app

from . import db
from .base import Base
from globus.manager.constants import ConfigurationObjectTag

from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import validates
from sqlalchemy.types import TypeDecorator


class EncryptedConfigurationType(TypeDecorator):
    """
    SQLAlchemy Type decorator that encrypts a column's value with the current
    flask app's keychain.
    """

    impl = db.TEXT

    def process_bind_param(self, value, dialect):
        if value:
            value = current_app.keychain.encrypt(value)
        else:
            value = "{}"
        return value

    def process_result_value(self, value, dialect):
        if value and value != "{}":
            value = MutableDict(current_app.keychain.decrypt(value))
        else:
            value = MutableDict()
        return value


class ConfigurationObject(Base):
    """
    ConfigurationObject

    The ConfigurationObject model is the parent of all models which are
    stored in an encrypted state in the GCS AppSync database. The table
    here mirrors the contents of the appsync database. Subclasses which
    are stored in the database must implement the following:

    - configuration_schema
        An object with a "dump" method that can process the model into a JSON
        object. This is used to update the configuration_hash column when the
        object is flushed to the database.
    - constructor
        The constructor for this object takes parameters for its columns:
        id, tag, serial, configuration_hash, and
        deleted. By default, this constructor uses the "tag" value (a
        ConfigurationObjectTag) as a discriminator for SQLalchemy's
        polymorphic object loading. An additional value may be passed as
        they secondary_discriminator parameter for objects where the tag plus
        some other data (most often a connector uuid) is needed to determine
        which class to use to load an object from the database.

        By convention, each subclass constructor accepts arguments of either
        appsync_data, or doc. appsync_data is used for data received
        from the appsync service that has already been validated by a DTN.
        It's a dict that contains the following values:

        -   owner_id: auth client id used to create this endpoint
        -   id: object id
        -   tag: one of the globus.manager.constants.ConfigurationObjectTag
            enum values
        -   serial: integer used to detect change conflicts--each change to
            the object in appsync increments this value.
        -   created_at: the time apspync created the object
        -   updated_at: the time apspync last updated the object
        -   configuration: encrypted data containing the object state. If this
            is "{}". there is no state yet or the object has been deleted.
        -   deleted: a boolean indicating that the object has been deleted in
            appsync.

        The doc argument is used when passing some sort of document containing
        the definition of the object. In many cases, this matches the values in
        the API schema, but may be different in some cases, based on the
        representation of the object. This is typically PUT into the object to
        set its state via properties, and then those proeprties are serialized
        to configuration when the object is prepared to be written to the
        database.

    Data encryption:

    The configuration_hash column stores a dict of data in the database. This
    data is encrypted at rest in the local database, using methods in the
    EncryptedConfigurationType, so properties that are stored in that are not
    easily queryable. For those cases, if the data needs to be queryied, use
    additional columns and indices in the object-type specific tables to handle
    that data.
    """

    __tablename__ = "configuration_object"
    __mapper_args__ = {"polymorphic_on": "discriminator"}

    # Database Columns
    id = db.Column("id", db.TEXT, primary_key=True, doc="ID of this object")
    tag = db.Column(
        "tag",
        db.Enum(ConfigurationObjectTag),
        nullable=False,
        doc="""
        The ConfigurationObjectTag used when sending this object to
        appsync. This allows us to get a rough idea of what is being stored
        there, without revealing too much about what the contents are.
        """,
    )
    serial = db.Column(
        db.INTEGER,
        nullable=False,
        doc="""
        The version number of this object. When an object is updated in
        appsync, this is checked to ensure that the new version is a change
        based on the previous version, and if not a conflict error is
        returned
        """,
    )
    discriminator = db.Column(
        "discriminator",
        db.TEXT,
        nullable=False,
        default="",
        server_default="",
        doc="""
        A string that allows SQLAlchemy determine which class to use when
        loading the data into the object. This is typically either the tag
        value or the tag with the UUID of the related connector.
        """,
    )
    configuration_hash = db.Column(
        "configuration",
        MutableDict.as_mutable(EncryptedConfigurationType),
        nullable=False,
        doc="""
        A dict containing the json document for this object. This is
        stored in the database as an encrypted binary blob in the
        "configuration" column.
        """,
    )
    _deleted = db.Column(
        "deleted",
        db.BOOLEAN,
        server_default="0",
        nullable=False,
        doc="""
        A flag indicating that an object has been soft deleted. In that case,
        the object's configuration may be the string `{}` to indicate that no
        data is available for the object any more.
        """,
    )

    @staticmethod
    def model_property(name, required: bool = False, doc: str = None):
        """
        This function acts as a replacement for @property to create a property
        which has its value within the current object's configuration_hash.
        Changes to the property update the configuration hash, and if the
        configuration hash has null value for the object it raises an
        AttributeError.

        :param name: The name of the property.
        :param required:
            A boolean flag. If set to True, the property can not be deleted. If
            False (the default) the property can be deleted.
        :param doc:
            An optional docstring for this property.
        :return:
            This function returns a property.
        """

        def fget(self):
            v = self.configuration_hash.get(name)
            if v is None:
                raise AttributeError(
                    "The {} property is not defined for this object".format(name)
                )
            return v

        def fset(self, value):
            if required and value is None:
                raise ValueError("The value for {} is invalid".format(name))
            old_value = self.configuration_hash.get(name)
            if value != old_value:
                if value is not None:
                    self.configuration_hash[name] = value
                else:
                    self.configuration_hash.pop(name, None)

        def fdel(self):
            if required:
                raise AttributeError("The {} property can not be deleted.".format(name))
            self.configuration_hash.pop(name, None)

        return property(fset=fset, fget=fget, fdel=fdel, doc=doc)

    @staticmethod
    def connector_property(name, required=False, doc=None):
        """
        Create a connector property on a connector's class allowing setting,
        getting, and deleting. These properties are managed in the "policies"
        subdocument of the model.
        """

        def fget(self):
            v = self.configuration_hash.get("policies", {}).get(name)
            if v is None:
                raise AttributeError(
                    "The {} property is not defined for this object".format(name)
                )
            return v

        def fset(self, value):
            if required and value is None:
                current_app.logger.debug(
                    "Clearing value for required connector property %s on %s",
                    name,
                    self,
                )

            if value != getattr(self, name, None):
                if "policies" not in self.configuration_hash:
                    self.configuration_hash["policies"] = MutableDict(
                        DATA_TYPE=self.POLICY_DATA_TYPE
                    )
                self.configuration_hash["policies"][name] = value
                self.configuration_hash.changed()

        def fdel(self):
            if required:
                current_app.logger.debug(
                    "Deleting value for required connector property %s on %s",
                    name,
                    self,
                )
            if (
                "policies" in self.configuration_hash
                and name in self.configuration_hash["policies"]
            ):
                del self.configuration_hash["policies"][name]

        return property(fset=fset, fget=fget, fdel=fdel, doc=doc)

    @classmethod
    def polymorphic_class(cls, discriminator, secondary=""):
        """
        Determines which model subclass constructor to use when given an
        encrypted blob, based on some discriminator value.

        TODO: currently, the discriminator is obtained by concatenating the TAG
        value which is stored in AppSync with an optional secondary
        discriminator. The secondary discriminator is used to choose when
        multiple subclasses implement different versions of the model. However,
        the value of the discriminator is not currently stored in the appsync
        schema, so is not available. It might be simpler to include a new field
        in the appsync schema and use that (uuid?) to choose which constructor
        to use to determine the class.
        """
        # Is there API for this?
        mapped = cls.__mapper__.polymorphic_map.get(
            ConfigurationObjectTag(discriminator).name + secondary
        )
        if mapped:
            return mapped.class_

    # Methods and properties for this class
    def __init__(self, id=None, tag=None, **kwargs):

        assert isinstance(tag, ConfigurationObjectTag)
        tag_string = tag.name

        secondary_discriminator = kwargs.pop("secondary_discriminator", "")

        if self.configuration_hash is None:
            self.configuration_hash = MutableDict()

        if "configuration_hash" in kwargs:
            kwargs["configuration_hash"] = MutableDict(**kwargs["configuration_hash"])

        super().__init__(
            id=id, tag=tag, discriminator=tag_string + secondary_discriminator, **kwargs
        )

        if self.id is None:
            if self.configuration_hash.get("id"):
                self.id = self.configuration_hash["id"]
            else:
                self.id = str(uuid.uuid4())
        if self.serial is None:
            self.serial = 0

    @validates("id")
    def immutable_properties(self, key, value):
        """
        Ensure that the "id" property is only ever set once on this object.
        """
        old_value = getattr(self, key, None)
        if isinstance(value, uuid.UUID):
            value_str = str(value)
        else:
            value_str = value
        if old_value is not None and old_value != value_str:
            # TODO: specialize the error for violation?
            raise ValueError("The {} property is immutable".format(key))
        return value_str

    def update(self, new_state):
        assert self.tag == ConfigurationObjectTag(new_state["tag"])
        self.created_at = new_state["created_at"]
        self.updated_at = new_state["updated_at"]
        self.serial = new_state["serial"]
        # Set our hard-delete value first so when we set configuration, we don't
        # overwrite the soft-delete value
        self.deleted = new_state["deleted"]
        self.configuration = new_state["configuration"]
        if self.configuration_hash.get("deleted"):
            self.deleted = True

        # The reinitializer has some restrictions on what it can do as far as
        # mutating columns and having them reliably be reflected on the next
        # object flush, so in general, changes from init_on_load should not
        # update any database columns.
        #
        # For some models that have extra columns which might be
        # changed in addition to the configuration_hash (for unique constraints
        # or foreign keys), there can be init_appsync() method that will do the
        # updates based on the document, but then also update the columns to
        # reflect the contents of the data.
        if hasattr(self, "init_appsync"):
            self.init_appsync()
        elif hasattr(self, "init_on_load"):
            self.init_on_load()

    def delete(self, *args, **kwargs):
        self.deleted = True

    @property
    def configuration(self):
        """
        A binary blob containing the encrypted data from appsync. This data
        can only be encrypt4ed by retrieving the DataKey object from
        appsync.
        """
        self.refresh_configuration_hash()

        if self.configuration_hash:
            return current_app.keychain.encrypt(self.configuration_hash)
        else:
            return "{}"

    @configuration.setter  # type: ignore
    def configuration(self, value):
        if value == "{}":
            self.configuration_hash = MutableDict()
        else:
            self.configuration_hash = MutableDict(**current_app.keychain.decrypt(value))
            if self.configuration_hash and self.configuration_hash.get("deleted"):
                self.deleted = True

    def raw_configuration(self):
        if self.configuration_hash:
            return current_app.keychain.encrypt(self.configuration_hash)
        else:
            return "{}"

    @hybrid_property
    def deleted(self):
        return bool(self._deleted)

    @deleted.expression  # type: ignore
    def deleted(cls):
        return cls._deleted

    @deleted.setter  # type: ignore
    def deleted(self, value):
        if value:
            self._deleted = value
        else:
            self._deleted = False
        # The refresh_configuration won't update the configuration_hash after we
        # mark ourselves as deleted, so do it here
        self.configuration_hash["deleted"] = self._deleted

    def write_config(self, delete: bool = False):
        """
        Models override this to have config files persisted to the filesystem
        for other services to find.

        :param delete:
            If True, this model is being removed from the system and its local
            filesystem config should also be deleted.
        """
        pass

    def refresh_configuration_hash(self):
        if not self.deleted:
            new_hash = self.configuration_schema.dump(self)
            if new_hash != self.configuration_hash:
                self.configuration_hash = MutableDict(**new_hash)

    def __repr__(self):
        return "<ConfigurationObject {} {}>".format(self.tag.name, self.id)
