import os
import pwd
import ssl
import stat
from typing import Optional

from flask import current_app
from OpenSSL import crypto

from globus.manager.api.schemas.api.domain import domain_private_schema, DomainDict
from globus.manager.utils.run_as_user import run_as_user
import globus.manager.utils.crypto_utils as crypto_utils


class Domain:
    """
    Model for a custom domain. This domain can be associated either with an
    endpoint or a collection. The admin provides the domain name, certififcate
    chain, and private key and GCS uses those values. The admin is responsible
    for handling the DNS updates when the DTN list changes.

    This is NOT a separate SQLAlchemy model. The value is embedded in the
    Endpoint or Collection model it is related to.
    """

    domain_dir = os.path.expanduser("~gcsweb/etc/domains")

    _DATA_TYPES = ["domain#1.0.0"]
    _DATA_TYPE = _DATA_TYPES[0]
    configuration_schema = domain_private_schema
    DOMAIN_PROPERTIES = {
        "domain#1.0.0": [
            "domain_name",
            "wildcard",
            "certificate",
            "certificate_chain",
            "private_key",
            "certificate_path",
            "certificate_chain_path",
            "private_key_path",
        ]
    }

    def __init__(self, appsync_data=None, doc: Optional[DomainDict] = None):
        self._certificate = None  # type: Optional[str]
        self._certificate_chain = None  # type: Optional[str]
        self._private_key = None  # type: Optional[str]
        self._certificate_path = None  # type: Optional[str]
        self._certificate_chain_path = None  # type: Optional[str]
        self._private_key_path = None  # type: Optional[str]
        self._wildcard = False

        if appsync_data is not None:
            self._put(**appsync_data)
        elif doc is not None and doc.get("DATA_TYPE") in self._DATA_TYPES:
            self.put({k: v for k, v in doc.items()})
            self.managed = hasattr(self, "certificate") and hasattr(self, "private_key")
        else:
            raise ValueError("Unsupported DATA_TYPE")

    @property
    def domain_name(self) -> str:
        return self._domain_name

    @domain_name.setter
    def domain_name(self, value: str):
        self._domain_name = value

    @property
    def wildcard(self) -> bool:
        return self._wildcard

    @wildcard.setter
    def wildcard(self, value: bool):
        self._wildcard = value

    @property
    def certificate(self) -> str:
        if self._certificate is not None:
            return self._certificate
        else:
            raise AttributeError(
                "The certificate property is not defined for this domain"
            )

    @certificate.setter
    def certificate(self, value: Optional[str]):
        if value is not None:
            self._certificate = value
        else:
            self._certificate = None

    @property
    def certificate_chain(self) -> str:
        if self._certificate_chain is not None:
            return self._certificate_chain
        else:
            raise AttributeError(
                "The certificate_chain property is not defined for this domain"
            )

    @certificate_chain.setter
    def certificate_chain(self, value: Optional[str]):
        if value is not None:
            self._certificate_chain = value
        else:
            self._certificate_chain = None

    @property
    def private_key(self) -> str:
        if self._private_key is not None:
            return self._private_key
        else:
            raise AttributeError(
                "The private_key property is not defined for this domain"
            )

    @private_key.setter
    def private_key(self, value: Optional[str]):
        if value is not None:
            self._private_key = value
        else:
            self._private_key = None

    @property
    def certificate_path(self) -> str:
        if self._certificate_path is not None:
            return self._certificate_path
        else:
            raise AttributeError(
                "The certificate_path property is not defined for this domain"
            )

    @certificate_path.setter
    def certificate_path(self, value: Optional[str]):
        if value is not None:
            self._certificate_path = value
        else:
            self._certificate_path = None

    @property
    def certificate_chain_path(self) -> str:
        if self._certificate_chain_path is not None:
            return self._certificate_chain_path
        else:
            raise AttributeError(
                "The certificate_chain_path property is not defined for this domain"
            )

    @certificate_chain_path.setter
    def certificate_chain_path(self, value: Optional[str]):
        if value is not None:
            self._certificate_chain_path = value
        else:
            self._certificate_chain_path = None

    @property
    def private_key_path(self) -> str:
        if self._private_key_path is not None:
            return self._private_key_path
        else:
            raise AttributeError(
                "The private_key_path property is not defined for this domain"
            )

    @private_key_path.setter
    def private_key_path(self, value: Optional[str]):
        if value is not None:
            self._private_key_path = value
        else:
            self._private_key_path = None

    @run_as_user(user="gcsweb")
    def write_config(self, delete=False):
        domain_cert_dir = os.path.join(self.domain_dir, self.domain_name)

        if not os.path.exists(domain_cert_dir):
            os.mkdir(domain_cert_dir, 0o0750)

        if not self.managed:
            # Create symlinks to the domain dir so gridftp can find them

            domain_cert_link = os.path.join(domain_cert_dir, "cert.pem")
            domain_cert_chain_link = os.path.join(domain_cert_dir, "chain.pem")
            domain_key_link = os.path.join(domain_cert_dir, "privkey.pem")

            if os.path.lexists(domain_cert_link):
                os.remove(domain_cert_link)
            os.symlink(self.certificate_path, domain_cert_link)

            if os.path.lexists(domain_cert_chain_link):
                os.remove(domain_cert_chain_link)

            if self._certificate_chain_path:
                os.symlink(self.certificate_chain_path, domain_cert_chain_link)
            if os.path.lexists(domain_key_link):
                os.remove(domain_key_link)
            os.symlink(self.private_key_path, domain_key_link)

            try:
                self.check_cert_permissions(self.certificate_path)
                if self._certificate_chain_path:
                    self.check_cert_permissions(self.certificate_chain_path)
                self.check_key_permissions(self.private_key_path)
            except ValueError as e:
                current_app.logger.error(
                    "Error configuring manually deployed certificate or key: %s", str(e)
                )
            return

        def write_to_file(contents, path, perms, mode):
            if os.path.lexists(path):
                os.remove(path)

            with open(os.open(path, mode, perms), "w") as f:
                f.write(contents)

        def _persist_to_filesystem(delete: bool = False) -> None:
            """
            Persist a managed certificate to the filesystem.
            """
            cert_perms = 0o644
            cert_mode = os.O_CREAT | os.O_WRONLY
            key_perms = 0o600
            key_mode = os.O_CREAT | os.O_WRONLY

            if delete:
                if os.path.lexists(self.certificate_path):
                    os.remove(self.certificate_path)
                if hasattr(self, "certificate_chain_path") and os.path.lexists(
                    self.certificate_chain_path
                ):
                    os.remove(self.certificate_chain_path)
                if os.path.lexists(self.private_key_path):
                    os.remove(self.private_key_path)
                try:
                    os.rmdir(os.path.join(self.domain_dir, self.domain_name))
                except Exception:
                    pass

                return

            write_to_file(
                self.certificate,
                self.certificate_path,
                cert_perms,
                cert_mode,
            )

            if hasattr(self, "certificate_chain_path"):
                write_to_file(
                    self.certificate_chain,
                    self.certificate_chain_path,
                    cert_perms,
                    cert_mode,
                )
            write_to_file(self.private_key, self.private_key_path, key_perms, key_mode)

        _persist_to_filesystem(delete=delete)

    @staticmethod
    def check_cert_permissions(path: str) -> None:
        try:
            valid_uid = [0, pwd.getpwnam("gcsweb").pw_uid]
            stat_res = os.stat(path)
            if stat_res.st_uid not in valid_uid:
                raise ValueError(
                    "The certificate file {} is not owned by root or the gcsweb "
                    "user".format(path)
                )
            if stat_res.st_mode & (
                stat.S_IXUSR | stat.S_IWGRP | stat.S_IXGRP | stat.S_IWOTH | stat.S_IXOTH
            ):
                raise ValueError(
                    "The permissions on the certificate file {} are too permissive "
                    "for use with Globus. The file must be readable and optionally "
                    "writable by the owner.".format(path)
                )
            if stat_res.st_size == 0:
                raise ValueError("The certificate file {} is empty.".format(path))
            if (stat_res.st_mode & stat.S_IFREG) == 0:
                raise ValueError(
                    "The certificate file {} is not a regular file.".format(path)
                )
        except Exception as e:
            current_app.logger.error(
                "Error with certificate permissions: {}".format(str(e))
            )
            raise ValueError(
                "The certificate file {} is either in a directory that "
                "is inaccessible by Globus, or does not exist".format(path)
            )

    @staticmethod
    def check_key_permissions(path: str) -> None:
        try:
            valid_uid = [0, pwd.getpwnam("gcsweb").pw_uid]
            stat_res = os.stat(path)
            if stat_res.st_uid not in valid_uid:
                raise ValueError(
                    "The certificate file {} is not owned by root or the gcsweb "
                    "user".format(path)
                )
            if stat_res.st_mode & (
                stat.S_IXUSR
                | stat.S_IRGRP
                | stat.S_IWGRP
                | stat.S_IXGRP
                | stat.S_IROTH
                | stat.S_IWOTH
                | stat.S_IXOTH
            ):
                raise ValueError(
                    "The permissions on the certificate file {} are too permissive "
                    "for use with Globus. The file must be read only by the owner.".format(
                        path
                    )
                )
            if stat_res.st_size == 0:
                raise ValueError("The certificate file {} is empty.".format(path))
            if (stat_res.st_mode & stat.S_IFREG) == 0:
                raise ValueError(
                    "The certificate file {} is not a regular file.".format(path)
                )
        except Exception as e:
            current_app.logger.error(
                "Error with certificate permissions: {}".format(str(e))
            )
            raise ValueError(
                "The certificate file {} is either in a directory that "
                "is inaccessible by Globus, or does not exist".format(path)
            )

    def put(self, doc):
        DATA_TYPE = doc.get("DATA_TYPE")

        if DATA_TYPE not in self._DATA_TYPES:
            raise ValueError("The DATA_TYPE {} is not supported".format(DATA_TYPE))

        properties = self.DOMAIN_PROPERTIES[DATA_TYPE]
        for prop in doc:
            if prop == "DATA_TYPE":
                continue
            elif prop == "domain_name":
                if not isinstance(doc[prop], str) or "/" in doc[prop]:
                    raise ValueError("Invalid value for domain_name")
            if prop not in properties:
                raise ValueError("The {} property is not supported".format(prop))

        certificate = doc.get("certificate")
        certificate_chain = doc.get("certificate_chain")
        private_key = doc.get("private_key")

        certificate_path = doc.get("certificate_path")
        certificate_chain_path = doc.get("certificate_chain_path")
        private_key_path = doc.get("private_key_path")

        managed = bool(certificate and private_key)
        unmanaged = bool(certificate_path and private_key_path)

        if (managed and unmanaged) or not (managed or unmanaged):
            raise ValueError("Invalid {} document".format(DATA_TYPE))

        if (managed and certificate_chain_path) or (unmanaged and certificate_chain):
            raise ValueError("Invalid {} document".format(DATA_TYPE))

        _doc = doc.copy()
        if unmanaged:
            self.check_cert_permissions(certificate_path)
            self.check_key_permissions(private_key_path)
            if certificate_chain_path:
                self.check_cert_permissions(certificate_chain_path)
        else:
            try:
                _doc["certificate"] = crypto_utils.parse_certificate(
                    doc["certificate"], hostname=doc["domain_name"]
                )
            except crypto.Error as e:
                current_app.logger.error("Unable to parse certificate data: %s", str(e))
                raise ValueError("Unable to parse the certificate data")
            except ssl.CertificateError as e:
                current_app.logger.error(str(e))
                raise ValueError("The certificate does match the requested domain name")
            except Exception as e:
                current_app.logger.error(
                    "Unable to parse certificate data: %s %s", type(e), str(e)
                )
                raise ValueError("Invalid value for the certificate property.")

            if doc.get("certificate_chain"):
                # Parse these here to make sure that (1) they are legitimate PEM
                # data, and (2) remove any non-PEM data in the managed certificate
                try:
                    chain = crypto_utils.parse_certificate_chain(
                        doc["certificate_chain"]
                    )
                    if chain:
                        _doc["certificate_chain"] = chain
                except Exception as e:
                    current_app.logger.error(
                        "Unable to parse certificate_chain data: %s", str(e)
                    )
                    raise ValueError(
                        "Invalid value for the certificate_chain property."
                    )
            try:
                _doc["private_key"] = crypto_utils.parse_private_key(doc["private_key"])
            except Exception as e:
                current_app.logger.error("Unable to parse private_key data: %s", str(e))
                raise ValueError("Invalid value for the private_key property.")

        self._put(**_doc)

    def _put(self, DATA_TYPE=None, **doc):
        properties = self.DOMAIN_PROPERTIES[DATA_TYPE]

        for prop in properties:
            if prop in doc:
                setattr(self, prop, doc[prop])

        self.managed = hasattr(self, "certificate") and hasattr(self, "private_key")

        if self.managed:
            # If this is a managed certificate/key, then we'll need to at times
            # persist these to the filesystem so that apache can use them.
            self.certificate_path = os.path.join(
                self.domain_dir, "{}/cert.pem".format(self.domain_name)
            )
            self.private_key_path = os.path.join(
                self.domain_dir, "{}/privkey.pem".format(self.domain_name)
            )
            if hasattr(self, "certificate_chain"):
                self.certificate_chain_path = os.path.join(
                    self.domain_dir, "{}/chain.pem".format(self.domain_name)
                )

    def _patch(self, DATA_TYPE=None, **doc):
        properties = self.DOMAIN_PROPERTIES[DATA_TYPE]

        for prop in properties:
            if prop in doc:
                setattr(self, prop, doc[prop])

    def patch(self, **doc):
        DATA_TYPE = doc.get("DATA_TYPE")

        if DATA_TYPE not in self._DATA_TYPES:
            raise ValueError("The DATA_TYPE {} is not supported".format(DATA_TYPE))

        properties = self.DOMAIN_PROPERTIES[DATA_TYPE]
        for prop in doc:
            if prop == "DATA_TYPE":
                continue
            elif prop == "domain_name":
                if not isinstance(doc[prop], str) or "/" in doc[prop]:
                    raise ValueError("Invalid value for domain_name")
            if prop not in properties:
                raise ValueError("The {} property is not supported".format(prop))

        certificate = doc.get("certificate")
        certificate_chain = doc.get("certificate_chain")
        private_key = doc.get("private_key")

        certificate_path = doc.get("certificate_path")
        certificate_chain_path = doc.get("certificate_chain_path")
        private_key_path = doc.get("private_key_path")

        # Don't allow change from unmanaged to managed or vice versa via patch
        if self.is_managed and (
            certificate_path or certificate_chain_path or private_key_path
        ):
            raise ValueError("Invalid {} document".format(DATA_TYPE))
        if (not self.is_managed) and (certificate or certificate_chain or private_key):
            raise ValueError("Invalid {} document".format(DATA_TYPE))

        _doc = doc.copy()
        if not self.is_managed:
            if certificate_path:
                self.check_cert_permissions(certificate_path)
            if private_key_path:
                self.check_key_permissions(private_key_path)
            if certificate_chain_path:
                self.check_cert_permissions(certificate_chain_path)
        else:
            if doc.get("certificate"):
                try:
                    domain_name = doc.get("domain_name", self.domain_name)
                    _doc["certificate"] = crypto_utils.parse_certificate(
                        doc["certificate"], hostname=domain_name
                    )
                except crypto.Error as e:
                    current_app.logger.error(
                        "Unable to parse certificate data: %s", str(e)
                    )
                    raise ValueError("Unable to parse the certificate data")
                except ssl.CertificateError as e:
                    current_app.logger.error(str(e))
                    raise ValueError(
                        "The certificate does match the requested domain name"
                    )
                except Exception as e:
                    current_app.logger.error(
                        "Unable to parse certificate data: %s %s", type(e), str(e)
                    )
                    raise ValueError("Invalid value for the certificate property.")

            if doc.get("certificate_chain"):
                # Parse these here to make sure that (1) they are legitimate PEM
                # data, and (2) remove any non-PEM data in the managed certificate
                try:
                    chain = crypto_utils.parse_certificate_chain(
                        doc["certificate_chain"]
                    )
                    if chain:
                        _doc["certificate_chain"] = chain
                except Exception as e:
                    current_app.logger.error(
                        "Unable to parse certificate_chain data: %s", str(e)
                    )
                    raise ValueError(
                        "Invalid value for the certificate_chain property."
                    )
            if doc.get("private_key"):
                try:
                    _doc["private_key"] = crypto_utils.parse_private_key(
                        doc["private_key"]
                    )
                except Exception as e:
                    current_app.logger.error(
                        "Unable to parse private_key data: %s", str(e)
                    )
                    raise ValueError("Invalid value for the private_key property.")

        self._patch(**_doc)

    @property
    def files_exist(self):
        try:
            if self._certificate_path:
                self.check_cert_permissions(self.certificate_path)
            if self._private_key_path:
                self.check_key_permissions(self.private_key_path)
            if self._certificate_chain_path:
                self.check_cert_permissions(self.certificate_chain_path)
            return True
        except ValueError as e:
            current_app.logger.error("Misconfigured: %s", str(e))
            return False

    @property
    def is_managed(self):
        return bool(self._certificate and self._private_key)

    def __repr__(self):
        return "<Domain {}>".format(self.domain_name)
