import ipaddress

import marshmallow as ma
from marshmallow_oneofschema import OneOfSchema

from globus.manager.constants import NodeStatus
from .data_type import DataTypeSchema
from .schema_version_checker import version_safe_type_schemas
from ..patch_schema import create_patch_schema
from ..enum_field import EnumField


class NodeSchema_1_0_0(ma.Schema):
    _DATA_TYPE = "node#1.0.0"
    DATA_TYPE = ma.fields.String(required=True, allow_none=False, default=_DATA_TYPE)
    id = ma.fields.UUID()
    ip_addresses = ma.fields.List(ma.fields.String, required=True)

    incoming_port_range = ma.fields.List(
        ma.fields.Int(validate=[ma.validate.Range(min=0, max=2**16 - 1)]),
        validate=[ma.validate.Length(equal=2)],
        required=False,
    )

    outgoing_port_range = ma.fields.List(
        ma.fields.Int(validate=[ma.validate.Range(min=0, max=2**16 - 1)]),
        validate=[ma.validate.Length(equal=2)],
        required=False,
    )

    status = EnumField(enum_class=NodeStatus, required=True)

    @ma.validates("ip_addresses")
    def validate_ip_addresses(self, data, **kwargs):
        if data:
            try:
                for i in data:
                    ipaddress.ip_address(i)
            except Exception as e:
                raise ma.ValidationError(str(e))


class NodeSchema_1_1_0(NodeSchema_1_0_0):
    _DATA_TYPE = "node#1.1.0"
    DATA_TYPE = ma.fields.String(required=True, allow_none=False, default=_DATA_TYPE)

    data_interface = ma.fields.String(required=False, allow_none=True)

    @ma.validates("data_interface")
    def validate_data_interface(self, data, **kwargs):
        if data:
            try:
                ipaddress.ip_address(data)
            except Exception as e:
                raise ma.ValidationError(str(e))


node_1_0_0_schema = NodeSchema_1_0_0()
node_1_0_0_patch_schema = create_patch_schema(NodeSchema_1_0_0)

node_1_1_0_schema = NodeSchema_1_1_0()
node_1_1_0_patch_schema = create_patch_schema(NodeSchema_1_1_0)


class NodeSchema(OneOfSchema):
    _DATA_TYPE = node_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = version_safe_type_schemas(
        {"node#1.0.0": node_1_0_0_schema, "node#1.1.0": node_1_1_0_schema}
    )
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE


class NodePatchSchema(OneOfSchema):
    _DATA_TYPE = node_1_1_0_patch_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = version_safe_type_schemas(
        {"node#1.0.0": node_1_0_0_patch_schema, "node#1.1.0": node_1_1_0_patch_schema}
    )
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE


node_schema = NodeSchema()
node_patch_schema = NodePatchSchema(partial=True)

DataTypeSchema.register_schema(node_1_0_0_schema)
DataTypeSchema.register_schema(node_1_1_0_schema)
