from typing import List

import marshmallow as ma

from .data_type import DataTypeSchema, data_type_schema


class StringOrObjectField(ma.fields.Field):
    """
    Polymorphic field. This can be used ot de/serialize any object based on the
    DATA_TYPE field or a plain string.
    """

    def _serialize(self, value, attr, obj, **kwargs):
        dt = ma.utils.get_value(value, "DATA_TYPE")
        if dt:
            schema = data_type_schema.schema_for(dt)
            return schema.dump(value)
        else:
            return value

    def _deserialize(self, value, attr, data, **kwargs):
        dt = ma.utils.get_value(value, "DATA_TYPE")
        if dt:
            schema = data_type_schema.schema_for(dt)
            return schema.load(value)
        else:
            return value


class ResultSchema(ma.Schema):
    """
    This is the result envelope returned from all operations in this API. Each
    operation may add properties to this base document type with additional
    operation-specific data values.
    """

    _DATA_TYPE = "result#1.0.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    http_response_code = ma.fields.Int(
        required=True, validate=[ma.validate.Range(100, 599)]
    )
    code = ma.fields.String(required=True)
    detail = StringOrObjectField()
    message = ma.fields.String()
    data = ma.fields.Raw()
    has_next_page = ma.fields.Boolean(default=False)
    marker = ma.fields.String(allow_none=True)
    private = []  # type: List[str]


result_schema = ResultSchema()
DataTypeSchema.register_schema(result_schema)
