import datetime
import sched
from typing import Any, Dict, List

import flask

import globus_sdk
import globus.manager.api.models as models
import globus.manager.api.errors as errors
from globus.manager.api.blueprints.utils import (
    delete_object,
    update_object,
)
from .exclusive_subscription_lock import ExclusiveSubscriptionLock


class PendingUpdatesManager:
    """
    This class handles performing updates on objects related to API changes
    """

    PENDING_UPDATE_LIMIT = 5  # number of updates that are processed per poll
    PENDING_UPDATE_DELAY = 5  # delay between polls of the pending update list

    def __init__(
        self,
        scheduler: sched.scheduler,
        app: flask.Flask,
        subscription_lock: ExclusiveSubscriptionLock,
    ):
        """
        Initialize a PendingUpdate object to process updates.

        :param sched.scheduler scheduler:
            A scheduler which this class will use to register timed events.
        :param GCSManagerApp app:
            An app which this class will use to interact with appsync, the
            database, locks and logs.
        :param: subscription_lock
            Lock to use when processing query results. This locks out the GCS
            manager process from processing requests that would interact with
            the database while we're processing changes that might make many
            changes to the database.
        """
        self.scheduler = scheduler  # Scheduler to register events with
        self.app = app  # Flask application this object is a part of
        self.__subscription_lock = subscription_lock
        self.pending_update_event = self.scheduler.enter(5, 0, self.check_pending)

    def check_pending(self) -> None:
        """
        Look at the pending_updates table. If there are entries there, process
        the changes.
        """

        with self.app.app_context(), self.__subscription_lock.exclusive():
            for pending in models.PendingUpdate.query.order_by(
                models.PendingUpdate.change_time
            ).limit(self.PENDING_UPDATE_LIMIT):
                if pending.update_type == models.PendingUpdate.DELETE:
                    self.process_pending_delete(pending)
                elif pending.update_type == models.PendingUpdate.UPDATE_DOMAIN:
                    self.process_pending_update_domain(pending)
                elif pending.update_type == models.PendingUpdate.PATCH:
                    self.process_pending_patch(pending)
                else:
                    self.app.logger.error(
                        "Unhandled pending change for %s %s",
                        pending.pending_object,
                        pending.update_type.value,
                    )
                    models.db.session.delete(pending)
                models.db.session.commit()
        self.pending_update_event = self.scheduler.enter(5, 0, self.check_pending)

    def process_pending_delete(self, pending: models.PendingUpdate) -> None:
        model = pending.pending_object
        extra_args: Dict[str, Any] = {}
        self.app.logger.debug(f"Processing pending delete of {model}")
        if isinstance(model, models.Role):
            if model.primary_role:
                model.primary_role = False
                model.refresh_configuration_hash()
            extra_args["collection_deleted"] = True
        elif isinstance(model, models.UserCredential):
            # A UserCredential may be marked for pending delete
            # when a StorageGateway is (soft) deleted. In this
            # case, the pending delete can only be processed after all
            # collections using that credential are hard deleted, for
            # db consistency.
            #
            # To avoid blocking all progress if there are multiple
            # pending user credential deletes, update the time
            # stamp of the pending change.
            if (
                models.Collection.query.filter(
                    models.Collection.user_credential_id == model.id,
                    ~models.Collection.deleted,  # type: ignore
                ).count()
                > 0
            ):
                pending.change_time = datetime.datetime.now(tz=datetime.timezone.utc)
                models.db.session.add(pending)
                models.db.session.commit()
                return
        try:
            self.app.logger.debug(f"Deleting object {model} as a pending update")
            # Deleting the model will delete the pending object
            delete_object(model, **extra_args)
        except errors.InternalServerError as e:
            self.app.logger.error(e)
        except errors.APIBaseError as e:
            if e.status_code != 404:
                self.app.logger.error(e)

    def process_pending_update_domain(self, pending: models.PendingUpdate) -> None:
        self.app.logger.debug(f"Processing pending update of {pending.pending_object}")
        if isinstance(pending.pending_object, models.Collection):
            collection = pending.pending_object
            dependencies: List[models.ConfigurationObject] = [collection.endpoint]

            if (
                collection.is_guest_collection
                and not collection.is_deleted
                and pending.collection_serial is not None
            ):
                dependencies.append(collection.mapped_collection)
            try:
                update_object(
                    collection,
                    models.Collection.update_domain_name,
                    collection.domain_name,
                    dependent_objs=dependencies,
                    force=True,
                )
            except Exception as e:
                self.app.logger.error("Error updating %s: %s", collection, str(e))
        else:
            self.app.logger.error(
                f"Pending update domain for unsupported model type: {pending.pending_object}"
            )

        models.db.session.delete(pending)

    def _filter_guest_patch_doc(
        self,
        guest: models.Collection,
        full_patch_document: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Filter values from full_patch_document that would be illegal to apply given the mapped collection.
        """
        filtered_patch_document: Dict[str, Any] = {}

        # These fields can only be set on the guest when it is set to False on
        # the parent mapped collection.
        for k in ["force_encryption", "force_verify", "disable_verify"]:
            if (
                k in full_patch_document
                and getattr(guest.mapped_collection, k) is False
            ):
                filtered_patch_document[k] = full_patch_document[k]

        return filtered_patch_document

    def _patch_guest_collection(
        self,
        guest: models.Collection,
        patch_document: Dict[str, Any],
    ) -> None:
        """
        Patch the guest collection but filter out any changes that are invalid
        according to the latest version of the mapped collection.
        """
        filtered_patch_doc = self._filter_guest_patch_doc(guest, patch_document)
        if filtered_patch_doc:
            guest.patch(DATA_TYPE=models.Collection._DATA_TYPE, **patch_document)

    def process_pending_patch(self, pending: models.PendingUpdate) -> None:
        self.app.logger.debug(f"Processing pending patch of {pending.pending_object}")
        if isinstance(pending.pending_object, models.Collection):
            collection = pending.pending_object

            if not collection.is_deleted:
                guest_patch_doc: Dict[str, Any] = pending.data

                try:
                    if guest_patch_doc:
                        # Make sure that any updates that we are about to apply to a guest collection
                        # are still legal changes. We can remove changes from the docs that are no longer
                        # valid. However, it is still possible that we will submit the change to appsync
                        # and get a CONFLICT because the mapped collection has changed. In that case, it
                        # is possible that the updated mapped collection invalidates the guest update as
                        # well. Therefore, let's call update_object() instead of patch_object() with our
                        # own callable that mimics patch object but will filter based out invalid updates.

                        # We explicitly update the guest model which requires a model+appsync update.
                        # collection.patch(), called from self._patch_guest_collection(), will call
                        # update_transfer_object()
                        update_object(
                            collection,
                            lambda _: self._patch_guest_collection(
                                collection,
                                guest_patch_doc,
                            ),
                            dependent_objs=[
                                collection.mapped_collection,
                                collection.endpoint,
                            ],
                        )
                    else:
                        collection.update_transfer_endpoint()
                except globus_sdk.TransferAPIError:
                    pass  # Already calls log_transfer_exception(e)

        else:
            self.app.logger.error(
                f"Pending patch for unsupported model type: {pending.pending_object}"
            )

        models.db.session.delete(pending)
