import os
import json
import time
from typing import Tuple

import click
from flask import current_app
import requests

import globus.manager.api.models as models
from globus.manager.api.app import make_app
from globus.manager.scripts.utils.setup_utils import load_keychain
from globus.manager.utils.keychain import NodeKey


def oidc_register(
    ctx,
    display_name: str,
    discovery_url: str,
    domain: Tuple[str],
    support_contact: str,
    support_email: str,
    client_id: str,
    client_secret: str,
    username_claim: str,
    id_claim: str,
) -> None:
    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        ctx.exit(1)

    app = make_app(app_config="globus.manager.api.config.oidc_cli")
    models.db.init_app(app)

    with app.app_context():
        models.db.create_all()
        models.db.session.commit()

        # Needed by AppSync
        for conf in models.Configuration.query:
            if conf.name == "node_key":
                value = NodeKey(**json.loads(conf.value))
            else:
                value = conf.value
            app.config[conf.name.upper()] = value

        # Make sure `node setup` has been run
        if "CLIENT_ID" not in app.config:
            click.echo("This node has not been set up.", err=True)
            click.get_current_context().exit(1)

        try:
            # Confirm that the loaded configuration contains the client ID and secret,
            # otherwise this node has not been configured yet. Can also be checked in app.py:266.
            if not (app.config.keys() > {"CLIENT_ID", "CLIENT_SECRET"}):
                raise Exception(
                    "GCS node has not been configured. Please run 'node setup' "
                    "or run this command on an existing node."
                )

            # Load the keychain
            load_keychain()

            # Get the endpoint
            endpoint_model = models.Endpoint.query.one()
            if endpoint_model.has_oidc_server:
                click.echo(
                    "There is already a managed OIDC configuration for this endpoint, "
                    "please use the `oidc update` command to update parameters."
                )
                return
            elif not endpoint_model.has_oidc_registration:
                current_app.logger.debug("pre init")
                endpoint_model.oidc_registration = models.OIDCRegistration(
                    doc={
                        "DATA_TYPE": "oidc_registration#1.0.0",
                        "display_name": display_name,
                        "discovery_url": discovery_url,
                        "domains": None if not domain else list(domain),
                        "support_contact": support_contact,
                        "support_email": support_email,
                        "username_claim": username_claim,
                        "id_claim": id_claim,
                        "oidc_client_id": client_id,
                        "oidc_client_secret": client_secret,
                    }
                )
                current_app.logger.debug("init done")
                endpoint_model.refresh_configuration_hash()
                current_app.appsync_client.update_object(endpoint_model)
                models.db.session.add(endpoint_model)
                models.db.session.commit()

            try:
                idp_data = endpoint_model.oidc_registration.auth_identity_provider_data
                # FIXME: I can't tell if the idp data corresponds
                # to a registration for this domain for a different client ID
                # that got nuked
            except AttributeError:
                idp_data = None

            if not idp_data:
                # If the OIDC IDP hasn't been registered yet, we will try to do
                # so
                resp = None
                # Make sure the autodiscovery URL is reachable
                for _tries in range(10):
                    time.sleep(1)
                    try:
                        resp = requests.get(
                            endpoint_model.oidc_registration.discovery_url
                        )
                        if resp.status_code == 200:
                            break
                    except Exception as e:
                        err = "Error contacting discovery url: {}".format(str(e))
                        if _tries < 9:
                            current_app.logger.info(err)
                        else:
                            click.echo(err, err=True)
                            click.get_current_context().exit(1)

                if resp and resp.status_code == 200:
                    try:
                        endpoint_model.oidc_registration.register_idp()
                    except Exception as e:
                        click.echo("Error registering IdP: {}".format(str(e)), err=True)
                        click.get_current_context().exit(1)

                    endpoint_model.refresh_configuration_hash()
                    current_app.appsync_client.update_object(endpoint_model)
                    click.echo("The OIDC server has been registered for this endpoint.")
                    return

        except Exception as e:
            click.echo("Error while creating OIDC endpoint: {}".format(str(e)))
            click.get_current_context().exit(1)
